@@ -15,7 +15,6 @@
 """hardshare API client
 """
 import asyncio
-from concurrent.futures import FIRST_COMPLETED
 import hashlib
 import json
 import logging
@@ -220,7 +219,7 @@ class HSAPIClient:
                             await self.loop.sock_sendall(conn, bytes('ACTIVE:{}\n'.format(self.current.container_name), encoding='utf-8'))
 
                     elif msg == 'TERMINATE\n':
-                        print('received request: TERMINATE')
+                        logger.info('received request: TERMINATE')
                         main.cancel()
                         conn.close()
                         break
@@ -270,10 +269,11 @@ class HSAPIClient:
             self.main.cancel()
             self.loop.run_until_complete(self.main)
         except ValueError as e:
-            print(e)
+            logger.error('caught ValueError: {}'.format(e))
             return 1
-        self.dsocket.cancel()
-        self.loop.run_until_complete(self.dsocket)
+        finally:
+            self.dsocket.cancel()
+            self.loop.run_until_complete(self.dsocket)
         return 0
 
 
@@ -300,13 +300,16 @@ class HSAPIClient:
                 cprovider = self.current_wdeployment['cprovider']
                 cargs = self.current_wdeployment.get('cargs', None)
                 image = self.current_wdeployment.get('image', None)
-                self.current = core.WorkspaceInstance(cprovider=cprovider, cargs=cargs, image=image)
+                terminate = self.current_wdeployment.get('terminate', None)
+                init_inside = self.current_wdeployment.get('init_inside', [])
+                self.current = core.WorkspaceInstance(cprovider=cprovider, cargs=cargs, image=image, terminate=terminate)
                 self.loop.create_task(self.current.launch_instance(
                     instance_id=payload['id'],
                     ws_send=ws.send_str,
                     ws_recv=self.ws_recvmap[payload['id']],
                     conntype=payload['ct'],
                     initial_publickey=payload['pr'],
+                    init_inside=init_inside,
                     tunnelkey_path=self.local_config.get('ssh_key', None)
                 ))
                 logger.debug('in response to INSTANCE_LAUNCH, sending ACK')
@@ -417,20 +420,25 @@ class HSAPIClient:
     async def run(self):
         logger.info('start to advertise wdeployment {}'.format(self.current_wdeployment['id']))
         headers = self._add_key_header()
-        if self.verify_certs:
-            session = aiohttp.ClientSession(headers=headers)
-        else:
-            conn = aiohttp.TCPConnector(verify_ssl=False)
-            session = aiohttp.ClientSession(connector=conn, headers=headers)
         uri = self.base_uri + '/ad/{}'.format(self.current_wdeployment['id'])
-        try:
-            async with session.ws_connect(uri, timeout=90.0, autoping=True) as ws:
-                async for msg in ws:
-                    if not (await self.handle_wsrecv(ws, msg)):
-                        break
+        active = True
+        while active:
+            if self.verify_certs:
+                session = aiohttp.ClientSession(headers=headers)
+            else:
+                conn = aiohttp.TCPConnector(verify_ssl=False)
+                session = aiohttp.ClientSession(connector=conn, headers=headers)
+            try:
+                async with session.ws_connect(uri, timeout=90.0, autoping=True) as ws:
+                    async for msg in ws:
+                        if not (await self.handle_wsrecv(ws, msg)):
+                            break
 
-        except asyncio.CancelledError:
-            pass
+            except asyncio.CancelledError:
+                active = False
 
-        finally:
-            await session.close()
+            except Exception as e:
+                logger.error('caught {}: {}'.format(type(e), e))
+
+            finally:
+                await session.close()
