@@ -24,16 +24,19 @@ import subprocess
 import sys
 
 import yaml
+from requests.exceptions import ConnectionError
 
 from .core import WorkspaceInstance
 from .mgmt import get_local_config, add_key, add_ssh_path, list_local_keys
 from .mgmt import find_wd, modify_local
 from .api import HSAPIClient
+from .err import Error as HSError
+from .addons import camera_main
 
 
 logger = logging.getLogger('hardshare')
 logger.setLevel(logging.WARNING)
-loghandler = logging.handlers.WatchedFileHandler(filename='hardshare_client.log', mode='a')
+loghandler = logging.handlers.WatchedFileHandler(filename='hardshare_client.log', mode='a', delay=True)
 loghandler.setLevel(logging.DEBUG)
 loghandler.setFormatter(logging.Formatter('%(name)s.%(funcName)s (%(levelname)s) (pid: {});'
                                           ' %(asctime)s ; %(message)s'
@@ -50,7 +53,22 @@ def get_config_with_index(id_prefix=None):
     if len(config['wdeployments']) == 0:
         print(('ERROR: no workspace deployment in local configuration.'))
         return config, None, 1
-    if id_prefix:
+    if isinstance(id_prefix, list):
+        if len(id_prefix) == 0:
+            if len(config['wdeployments']) > 1:
+                print('ERROR: ambiguous command: more than 1 workspace deployment defined.')
+                return config, None, 1
+            index = [0]
+        else:
+            indices = []
+            for idp in id_prefix:
+                index = find_wd(config, idp)
+                if index is None:
+                    print('ERROR: given prefix does not match precisely 1 workspace deployment')
+                    return config, None, 1
+                indices.append(index)
+            index = indices
+    elif id_prefix:
         index = find_wd(config, id_prefix)
         if index is None:
             print('ERROR: given prefix does not match precisely 1 workspace deployment')
@@ -159,6 +177,11 @@ def main(argv=None):
     register_parser = subparsers.add_parser('register',
                                             description=register_commanddesc,
                                             help=register_commanddesc)
+    register_parser.add_argument('--permit-more', action='store_false', default=True,
+                                 dest='register_at_most_one',
+                                 help=('permit registration of more than 1 wdeployment; '
+                                       'default is to fail if local configuration already '
+                                       'has wdeployment declared'))
 
     check_commanddesc = 'check registration of this workspace deployment'
     check_parser = subparsers.add_parser('check',
@@ -199,6 +222,28 @@ def main(argv=None):
                                   help='detach from invoking terminal (i.e., run as daemon)',
                                   dest='become_daemon')
 
+    attach_camera_commanddesc = 'attach camera stream to workspace deployments'
+    attach_camera_parser = subparsers.add_parser('attach-camera',
+                                                 description=attach_camera_commanddesc,
+                                                 help=attach_camera_commanddesc)
+    attach_camera_parser.add_argument('camera', default=0,
+                                      type=int,
+                                      help=('on Linux, 0 typically implies /dev/video0; '
+                                            'if you only have one camera, then try 0'))
+    attach_camera_parser.add_argument('id_prefix', metavar='ID', nargs='*', default=None,
+                                      help=('id of workspace deployment on which to attach'
+                                            ' (can be unique prefix); '
+                                            'this argument is not required '
+                                            'if there is only 1 workspace deployment'))
+    attach_camera_parser.add_argument('--width-height', metavar='W,H', type=str,
+                                      dest='attach_camera_res', default=None,
+                                      help=('width and height of captured images; '
+                                            'default depends on the supporting drivers'))
+    attach_camera_parser.add_argument('--crop', metavar='CROPCONFIG', type=str,
+                                      dest='attach_camera_crop_config', default=None,
+                                      help=('image crop configuration; '
+                                            'default: all wdeployments get full images'))
+
     terminate_commanddesc = 'mark as unavailable; optionally wait for current instance to finish'
     terminate_parser = subparsers.add_parser('terminate',
                                              description=terminate_commanddesc,
@@ -239,6 +284,8 @@ def main(argv=None):
                 dissolve_parser.print_help()
             elif argv_parsed.help_target_command == 'status':
                 status_parser.print_help()
+            elif argv_parsed.help_target_command == 'attach-camera':
+                attach_camera_parser.print_help()
             elif argv_parsed.help_target_command == 'ad':
                 advertise_parser.print_help()
             elif argv_parsed.help_target_command == 'terminate':
@@ -290,6 +337,36 @@ def main(argv=None):
         else:  # output_format == 'yaml'
             print(yaml.dump(findings, default_flow_style=False))
 
+
+    elif argv_parsed.command == 'attach-camera':
+        config, indices, rc = get_config_with_index(argv_parsed.id_prefix)
+        if rc != 0:
+            return rc
+        wdeployments = [config['wdeployments'][jj]['id'] for jj in indices]
+
+        local_keys = list_local_keys()
+        if len(local_keys) < 1:
+            print('No valid keys available. Check: `hardshare config -l`')
+            return 1
+        with open(local_keys[0], 'rt') as fp:
+            tok = fp.read().strip()
+
+        if argv_parsed.attach_camera_res:
+            width, height = [int(x) for x in argv_parsed.attach_camera_res.split(',')]
+            if width < 1 or height < 1:
+                print('Width, height must be positive')
+                return 1
+        else:
+            width, height = None, None
+
+        if argv_parsed.attach_camera_crop_config:
+            crop = json.loads(argv_parsed.attach_camera_crop_config)
+        else:
+            crop = None
+
+        camera_main(wdeployments, tok=tok, dev=argv_parsed.camera, width=width, height=height, crop=crop)
+
+
     elif argv_parsed.command == 'ad':
         if ac is None:
             print('cannot register without initial local configuration.'
@@ -347,7 +424,14 @@ def main(argv=None):
             print('cannot register without initial local configuration.'
                   ' (try `hardshare config --create`)')
             return 1
-        print(ac.register_new())
+        try:
+            print(ac.register_new(at_most_one=argv_parsed.register_at_most_one))
+        except HSError as err:
+            print('ERROR: {}'.format(err))
+            return 1
+        except ConnectionError:
+            print('ERROR: failed to reach server. Are you connected to the Internet?')
+            return 1
 
     elif argv_parsed.command == 'check':
         if ac is None:
@@ -373,7 +457,7 @@ def main(argv=None):
             print('\tcreated: {}'.format(res['date_created']))
             print('\torigin (address) of registration: {}'.format(res['origin']))
             if 'date_dissolved' in res:
-                print('\tdissolved: {}'.format(res['origin']))
+                print('\tdissolved: {}'.format(res['date_dissolved']))
 
     elif argv_parsed.command == 'dissolve':
         if ac is None:
@@ -420,6 +504,13 @@ def main(argv=None):
                     'remote': remote_config,
                 }
 
+            if 'local' in config:
+                ref = config['local']['wdeployments']
+            else:
+                ref = config['wdeployments']
+            for jj, wdeployment in enumerate(ref):
+                ref[jj]['url'] = 'https://rerobots.net/workspace/{}'.format(wdeployment['id'])
+
             if output_format == 'json':
                 print(json.dumps(config))
 
@@ -439,7 +530,7 @@ def main(argv=None):
                     for wdeployment in config['local']['wdeployments']:
                         print('{}\n\turl: {}\n\towner: {}\n\tcprovider: {}\n\tcargs: {}\n\timg: {}'.format(
                             wdeployment['id'],
-                            'https://rerobots.net/workspace/{}'.format(wdeployment['id']),
+                            wdeployment['url'],
                             wdeployment['owner'],
                             wdeployment['cprovider'],
                             wdeployment['cargs'],
